import inspect
import typing
from functools import wraps

from fastapi.openapi.models import OAuthFlowAuthorizationCode, OAuthFlows
from fastapi.security import OAuth2
from starlette.authentication import requires
from starlette.exceptions import HTTPException
from starlette.requests import Request

from .aad_options import AzureAdSettings
from .aad_user import AadUser


def authorize(
    scopes: typing.Union[str, typing.Sequence[str]] = None,
    roles: typing.Union[str, typing.Sequence[str]] = None,
):
    """authorize decorator. you can specify scopes and (or) roles"""

    def wrapper(endpoint):
        @wraps(endpoint)
        @requires(scopes)
        async def require_auth_endpoint(request: Request, *args, **kwargs):
            def has_required_roles(user_roles: typing.Sequence[str]) -> bool:
                for mandatory_role in mandatory_roles_list:
                    if mandatory_role not in user_roles:
                        return False
                return True

            # Check args
            mandatory_roles_list = []
            if roles is not None:
                mandatory_roles_list = (
                    [roles] if isinstance(roles, str) else list(roles)
                )

            # Since we used @requires and the caller has the Depends(aad.oauth2),
            # the request.user exists.
            user: AadUser = request.user
            user_roles_list = user.roles_id or []

            if len(mandatory_roles_list) > 0 and not has_required_roles(
                user_roles_list
            ):
                raise HTTPException(status_code=403, detail="Unauthorized role")

            if inspect.iscoroutinefunction(endpoint):
                result = await endpoint(request, *args, **kwargs)
            else:
                result = endpoint(request, *args, **kwargs)

            return result

        return require_auth_endpoint

    return wrapper


def oauth2_scheme(options: AzureAdSettings = None, env_path: str = None, **kwargs):
    """get the OAUTH2 schema used for API Authentication"""

    if options is None:
        options = AzureAdSettings(_env_file=env_path)

    scopes_dictionary = {}
    if options.scopes_list is not None:
        for ls in options.scopes_list:
            scopes_dictionary[ls] = ls

    return OAuth2(
        flows=OAuthFlows(
            authorizationCode=OAuthFlowAuthorizationCode(
                authorizationUrl=options.authorization_url,
                tokenUrl=options.token_url,
                scopes=scopes_dictionary,
            ),
        )
    )



import json
from typing import Dict, List, Optional, Union
from uuid import uuid4

import msal
from azure.core.exceptions import ClientAuthenticationError
from azure.identity import (
    AzureCliCredential,
    ChainedTokenCredential,
    ManagedIdentityCredential,
)

from .aad_autherror import AuthError
from .aad_cache_manager import CacheManager
from .aad_helpers import ensure_user_from_token
from .aad_options import AzureAdSettings
from .aad_token import AuthToken
from .aad_user import AadUser
from .aad_vault import AadVault


class AadClient:
    def __init__(
        self,
        session: dict = None,
        options: AzureAdSettings = None,
        env_path: str = None,
    ):

        if options is None:
            options = AzureAdSettings(_env_file=env_path)

        self.options = options
        self.session = session if session is not None else {}
        self.cache_manager = CacheManager

        # Session ID, representing the token STATE ID
        self.aad_id: str = self.session.get("aad_id", None)

        # On first auth, no token is cached, so no STATE ID
        if self.aad_id is None:
            self.aad_id = str(uuid4())

        self.session["aad_id"] = self.aad_id
        self.cache = msal.SerializableTokenCache()

    async def _load_cache(self):

        # get session cache
        cache_session = await self.cache_manager.get(self.aad_id)

        if "token_cache" in cache_session:
            self.cache.deserialize(cache_session["token_cache"])

        return self.cache

    async def _save_cache(self):
        if self.cache.has_state_changed:
            cache_session = await self.cache_manager.get(self.aad_id)
            cache_session["token_cache"] = self.cache.serialize()
            await self.cache_manager.set(self.aad_id, cache_session)

    def _ensure_options(self):
        if self.options is None:
            raise AuthError(
                "options_missing", "options are missing from aad authentication client"
            )

        def _ensure_options_property(property):
            if not hasattr(self.options, property):
                raise AuthError(
                    f"options_{property}_missing",
                    f"{property} is missing from options in aad authentication client",
                )

        _ensure_options_property("client_id")
        _ensure_options_property("authority")

        if not hasattr(self.options, "client_secret"):
            _ensure_options_property("vault_name")
            _ensure_options_property("vault_certificate_name")

    async def build_msal_public_app(self, **kwargs) -> msal.PublicClientApplication:
        self._ensure_options()
        token_cache = await self._load_cache()

        return msal.PublicClientApplication(
            client_id=self.options.client_id,
            authority=self.options.authority,
            token_cache=token_cache,
            **kwargs,
        )

    def _extract_access_token_from_msal_token(self, token: dict = None):

        if token is None:
            return None

        # extract token for web api scope only:
        target = self.options.scopes_list[0]

        token_values = list(token.values())
        for tok in token_values:
            if "target" in tok and tok["target"] == target:
                return tok["secret"]

        return token_values[0].get("secret", None) if len(token_values) > 0 else None

    async def get_user(self, validate: bool = False) -> Optional[AadUser]:

        cache_session = await self.cache_manager.get(self.aad_id)
        if not cache_session:
            return None

        token_cache = cache_session.get("token_cache", None)

        if not token_cache:
            return None

        msal_token = json.loads(token_cache)

        access_token = self._extract_access_token_from_msal_token(
            msal_token.get("AccessToken", None)
        )

        # Invalid. Gracefuly return
        if access_token is None:
            return None

        id_token = self._extract_access_token_from_msal_token(
            msal_token.get("IdToken", None)
        )
        refresh_token = self._extract_access_token_from_msal_token(
            msal_token.get("RefreshToken", None)
        )

        auth_token = AuthToken(
            access_token=access_token,
            id_token=id_token,
            refresh_token=refresh_token,
        )
        user = ensure_user_from_token(auth_token=auth_token, validate=validate)

        return user

    async def acquire_user_token(
        self,
        user: AadUser = None,
        scopes: Union[List[str], str] = None,
        **kwargs,
    ) -> Optional[AadUser]:

        try:
            self._ensure_options()

            authentication_token: AuthToken = None
            exception: Exception = None

            # Get scopes
            scopes = self.options.scopes_list if scopes is None else scopes
            scopes = scopes if isinstance(scopes, List) else [scopes]

            # ---------------------------------------------------
            # Case 0: Login fallback, only for login UI
            # ---------------------------------------------------
            # hiddens parameters to ensure simplicity of usage
            # these parameters are only used on user login, through web ui
            auth_response = kwargs.pop("auth_response", None)
            flow = kwargs.pop("flow", None)

            # particular case on login identication
            if auth_response is not None:
                aad_user = await self._acquire_token_and_user_by_auth_code_flow(
                    scopes, auth_response=auth_response, flow=flow, **kwargs
                )
                return aad_user

            # try to get a valid user from cache
            if user is None:
                user = await self.get_user(validate=False)

            # ---------------------------------------------------
            # Case 1: user is knwon
            # ---------------------------------------------------
            # we have a valid user in parameter with an existing token
            # if the audience from this issuer is the good one,
            # we can keep the token
            if user is not None and user.auth_token is not None:
                try:
                    authentication_token = await self._acquire_token_on_behalf_of(
                        user.auth_token, scopes, **kwargs
                    )
                except Exception as ex:
                    authentication_token = None
                    exception = Exception(ex, exception)

            # ---------------------------------------------------
            # Case 2: Try Managed Identity first or Azure CLI
            # ---------------------------------------------------
            if authentication_token is None:
                try:
                    authentication_token = (
                        await self._acquire_token_for_managed_identity(**kwargs)
                    )
                except Exception as ex:
                    authentication_token = None
                    exception = Exception(ex, exception)

            # ---------------------------------------------------
            # Case 3: Try to reach Service Principal directly
            # ---------------------------------------------------
            if authentication_token is None:
                try:
                    authentication_token = await self._acquire_token_for_client(
                        scopes, **kwargs
                    )
                except Exception as ex:
                    exception = Exception(ex, exception)
                    authentication_token = None

            if authentication_token is None:
                raise AuthError(exception=exception)

            # We can't validate the signature for Graph, since only Microsoft Graph can
            aad_user = ensure_user_from_token(
                auth_token=authentication_token, validate=True, **kwargs
            )

            return aad_user
        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def _acquire_token_and_user_by_auth_code_flow(
        self,
        scopes: Union[List[str], str],
        auth_response=None,
        flow=None,
        **kwargs,
    ) -> AadUser:
        try:
            self._ensure_options()

            cache_session = await self.cache_manager.get(self.aad_id)

            scopes = scopes if isinstance(scopes, List) else [scopes]

            cca = await self.build_msal_confidential_app(**kwargs)

            token = await self._get_token_from_cache(cca, scopes, **kwargs)

            if token:
                return ensure_user_from_token(auth_token=token, **kwargs)

            if flow:
                auth_code_flow = flow
            else:
                auth_code_flow = cache_session.get("flow", None)

            if auth_code_flow is None:
                raise AuthError(
                    "no_auth_code_flow",
                    (
                        "no auth code flow has been provided. Please be sure you have "
                        "a session set."
                    ),
                )

            token = cca.acquire_token_by_auth_code_flow(
                auth_code_flow=auth_code_flow, auth_response=auth_response, **kwargs
            )

            if not token or "access_token" not in token:
                error_code = "no_token_provided_from_acquire_token"
                error_desc = (
                    "can't get token from acquire token interactively "
                    "or from client browser"
                )
                error_code = (
                    error_code if token is None else token.get("error", error_code)
                )
                error_desc = (
                    error_desc
                    if token is None
                    else token.get("error_description", error_desc)
                )
                raise AuthError(error_code, error_desc)

            user = ensure_user_from_token(auth_token=AuthToken(token=token), **kwargs)

            await self._save_cache()

            return user

        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def _acquire_token_for_managed_identity(self, scope: str, **kwargs):
        try:
            # Managed identity when deployed,
            managed_identity = ManagedIdentityCredential()
            # Fallback on Azure CLI if running locally
            # The Azure CLI should be added to App Registration
            azure_cli = AzureCliCredential()

            credential = ChainedTokenCredential(managed_identity, azure_cli)

            # resource = (
            #     self.options.api_scopes_identifiers_root[0]
            #     if self.options.api_scopes_identifiers_root is not None
            #     and len(self.options.api_scopes_identifiers_root) >= 1
            #     else None
            # )

            # access_token = credential.get_token(resource)

            access_token = credential.get_token(scope)

            if access_token is None or access_token.token is None:
                raise AuthError(
                    "no_token_acquired_from_managed_identity",
                    "Can't get a token for the managed identity",
                )

            auth_token = AuthToken(access_token=access_token.token)

            return auth_token
        except ClientAuthenticationError as cae:
            raise AuthError(cae.error, cae.message)
        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def _acquire_token_for_client(
        self,
        scopes: Union[List[str], str],
        **kwargs,
    ):
        """Acquire a tocken to access a ressource from the application
        for specific scopes"""

        try:

            self._ensure_options()

            cca = await self.build_msal_confidential_app(**kwargs)

            scopes = scopes if isinstance(scopes, List) else [scopes]

            auth_token_from_cache = await self._get_token_from_cache(
                cca, scopes, **kwargs
            )

            if auth_token_from_cache:
                return auth_token_from_cache

            token = cca.acquire_token_for_client(scopes=scopes, **kwargs)

            if not token or "access_token" not in token:
                error_code = "no_token_provided_from_acquire_token_for_client"
                error_desc = "can't get token directly from application identity"
                error_code = (
                    error_code if token is None else token.get("error", error_code)
                )
                error_desc = (
                    error_desc
                    if token is None
                    else token.get("error_description", error_desc)
                )
                raise AuthError(error_code, error_desc)

            auth_token = AuthToken(token=token)

            await self._save_cache()

            return auth_token

        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def _acquire_token_on_behalf_of(
        self,
        auth_token: AuthToken,
        scopes: Union[List[str], str],
        **kwargs,
    ):
        try:

            self._ensure_options()

            cca = await self.build_msal_confidential_app(**kwargs)

            scopes = scopes if isinstance(scopes, List) else [scopes]
            auth_token_from_cache = await self._get_token_from_cache(
                cca, scopes, **kwargs
            )

            if auth_token_from_cache:
                return auth_token_from_cache

            token = cca.acquire_token_on_behalf_of(
                user_assertion=auth_token.access_token, scopes=scopes, **kwargs
            )

            if not token or "access_token" not in token:
                error_code = "no_token_provided_from_acquire_token_on_behalf_of"
                error_desc = "can't get token from acquire token on behalf of user"
                error_code = (
                    error_code if token is None else token.get("error", error_code)
                )
                error_desc = (
                    error_desc
                    if token is None
                    else token.get("error_description", error_desc)
                )
                raise AuthError(error_code, error_desc)

            auth_token_on_behalf_of = AuthToken(token=token)

            await self._save_cache()

            return auth_token_on_behalf_of

        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def build_msal_confidential_app(
        self, **kwargs
    ) -> msal.ConfidentialClientApplication:

        """Get an msal instance
        :return:
            The msal instance configured to use a certificate from the
            general Key Vault.
        """

        try:
            self._ensure_options()

            # get the client secret of any compatible client_credential from kwargs
            client_credential = kwargs.get(
                "client_credential", self.options.client_secret
            )

            # if no client secret, load a certificate
            if not client_credential:

                cache_session = await self.cache_manager.get("azure_ad")

                if cache_session is None or "certificate" not in cache_session:

                    aadVault = AadVault(self.options.vault_url)
                    certificate = await aadVault.get_certificate(
                        self.options.vault_certificate_key
                    )

                    cache_session = await self.cache_manager.get("azure_ad")
                    cache_session["certificate"] = certificate

                    await self.cache_manager.set("azure_ad", cache_session)

                # cache_session = self.cache_manager.read(self.aad_id)
                cache_session = await self.cache_manager.get("azure_ad")

                client_credential = aadVault.get_msal_client_credential(
                    cache_session["certificate"]
                )

            token_cache = await self._load_cache()

            # Create tha msal app from that
            return msal.ConfidentialClientApplication(
                client_id=self.options.client_id,
                authority=self.options.authority,
                client_credential=client_credential,
                token_cache=token_cache,
            )

        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def build_auth_code_flow(
        self,
        scopes: Union[List[str], str],
        redirect_uri=None,
        **kwargs,
    ):

        """Initiate an auth code flow.
        :return:
            The auth code flow containing the auth_uri you need to specify in
            your login uri
        """

        try:
            self._ensure_options()

            cache_session = await self.cache_manager.get(self.aad_id)

            scopes = scopes if isinstance(scopes, List) else [scopes]

            cca = await self.build_msal_confidential_app(**kwargs)

            flow: Dict[str, str] = cca.initiate_auth_code_flow(
                scopes, redirect_uri=redirect_uri
            )

            cache_session["flow"] = flow

            if self.session is not None:
                # save session
                self.session["aad_id"] = self.aad_id
                # save to cache
                await self.cache_manager.set(self.aad_id, cache_session)

            return flow

        except AuthError as aex:
            raise aex
        except Exception as ex:
            raise AuthError(exception=ex)

    async def _get_token_from_cache(
        self, cca: msal.ClientApplication, scopes: List[str], **kwargs
    ):
        accounts = cca.get_accounts()
        token = None
        auth_token = None

        if accounts:  # So all account(s) belong to the current signed-in user
            token = cca.acquire_token_silent(scopes, account=accounts[0], **kwargs)

            if not token or "access_token" not in token:
                return None

            auth_token = AuthToken(token=token)

            await self._save_cache()

        return auth_token

    def get_logout_uri(self, logout_page_uri):
        return (
            self.options.authority
            + "/oauth2/v2.0/logout"
            + "?post_logout_redirect_uri="
            + logout_page_uri
        )

    async def remove_account(self, **kwargs):

        try:
            self._ensure_options()
            # remove account from msal cache
            cca = await self.build_msal_confidential_app(**kwargs)

            accounts = cca.get_accounts()

            if accounts:
                cca.remove_account(accounts[0])

            await self._save_cache()

            # remove account from cache
            await self.cache_manager.remove(self.aad_id)

            # remove session id, stored in the cookie
            self.session.pop("aad_id")

        except Exception:
            pass  # do not raise any error here. just kindly exit the function



import os
import pathlib

import pytest
from async_asgi_testclient import TestClient
from fastapi import FastAPI
from fastapi.param_functions import Depends
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import Request

from aad import AadBearerBackend, authorize, oauth2_scheme

os.environ["CLIENT_ID"] = "01010101-aaaa-bbbb-acdf-020202020202"
os.environ["TENANT_ID"] = "02020202-aaaa-erty-olki-020202020202"
os.environ["DOMAIN"] = "contoso.onmicrosoft.com"
os.environ["API_SCOPES"] = "user_impersonation"
os.environ["AUTHORITY"] = "https://login.microsoftonline.com/common"
os.environ["VAULT_URL"] = "https://your_key_vault_name.vault.azure.net/"
os.environ["VAULT_CERTIFICATE_NAME"] = "certificate_name_from_your_key_vault"


@pytest.fixture(scope="module")
def public_key(cert):
    return cert["public_key"]


@pytest.fixture(scope="module")
def private_key(cert):
    return cert["private_key"]


@pytest.fixture(scope="module")
def thumbprint():
    return "ce496c5eff03728f9fd01b106874245ffc31058b".encode()


@pytest.fixture(scope="module")
def client_credential(private_key, public_key, thumbprint):
    class client_certificate:
        def __init__(self, private_key, public_key, thumbprint):
            self.private_key = private_key
            self.thumbprint = thumbprint
            self.public_key = public_key

    return client_certificate(private_key, public_key, thumbprint)


@pytest.fixture(scope="module")
def cert():

    # Get certificate private and public keys to create access_token
    aad_tests_dir = pathlib.Path(__file__).parent.absolute()

    public_key_file = os.path.join(aad_tests_dir, "public.pem")
    public_key = open(public_key_file, "r").read()

    private_key_file = os.path.join(aad_tests_dir, "private.pem")
    private_key = open(private_key_file, "r").read()

    return {"public_key": public_key, "private_key": private_key}


@pytest.fixture(scope="module")
def client(public_key):

    # pre fill client id
    swagger_ui_init_oauth = {
        "clientId": os.environ.get("CLIENT_ID"),
        "appName": "B-ID",
    }

    app = FastAPI(swagger_ui_init_oauth=swagger_ui_init_oauth)

    aad_bearer_backend = AadBearerBackend(public_key=public_key)
    # Add the bearer middleware
    app.add_middleware(AuthenticationMiddleware, backend=aad_bearer_backend)

    @app.get("/test")
    async def get_test(request: Request):
        return {"test": True}

    @app.get("/isauth")
    async def get_isauth(request: Request, token=Depends(oauth2_scheme())):
        return request.user

    @app.get("/isauth_impersonation")
    @authorize("user_impersonation")
    async def get_isauth_with_impersonation(
        request: Request, token=Depends(oauth2_scheme())
    ):
        return request.user

    @app.get("/isauth_impersonation_roles")
    @authorize("user_impersonation", ["Admin", "Contributor"])
    async def get_isauth_with_impersonation_and_roles(
        request: Request, token=Depends(oauth2_scheme())
    ):
        return request.user

    return TestClient(app)



import requests


class MinimalHttpClient:
    def __init__(self, verify=True, proxies=None, timeout=None):
        self.session = requests.Session()
        self.session.verify = verify
        self.session.proxies = proxies
        self.timeout = timeout

    def post(self, url, params=None, data=None, headers=None, **kwargs):
        return MinimalResponse(
            requests_resp=self.session.post(
                url, params=params, data=data, headers=headers, timeout=self.timeout
            )
        )

    def get(self, url, params=None, headers=None, **kwargs):
        return MinimalResponse(
            requests_resp=self.session.get(
                url, params=params, headers=headers, timeout=self.timeout
            )
        )


class MinimalResponse(object):  # Not for production use
    def __init__(self, requests_resp=None, status_code=None, text=None):
        self.status_code = status_code or requests_resp.status_code
        self.text = text or requests_resp.text
        self._raw_resp = requests_resp

    def raise_for_status(self):
        if self._raw_resp is not None:  # Turns out `if requests.response` won't work
            # cause it would be True when 200<=status<400
            self._raw_resp.raise_for_status()



from typing import List
import typing
from urllib.parse import urlparse

import requests
from pydantic.env_settings import BaseSettings
from pydantic.fields import Field


class AzureAdSettings(BaseSettings):
    """
    Represents the Azure AD Settings.
    client_id       : Application Resgistration app id (GUID)
    authority       : Authority url.
        ex: "https://login.microsoftonline.com/{client_id}"
    domain          : Azure tenant domain. "ex:microsoft.onmicrosoft.com"
    tenant_id       : Azure tenant id. (GUID)
    api_scopes      : API scopes list (separated by a blank space).
        ex:"user_impersonation"
    vault_url : vault url, containing the certificate to use.
        ex:"https://rgbertkvd10.vault.azure.net/"
    vault_certificate_name : certificate name, contained in vault. ex:"mycert"

    if not present, each value will be retrieved from environment variables:
    "CLIENT_ID", "AUTHORITY", "DOMAIN", "TENANT_ID", "API_SCOPES",
    "VAULT_URL", "VAULT_CERTIFICATE_NAME"

    """

    client_id: str = Field(None, description="Client id", env="CLIENT_ID")
    client_secret: str = Field(None, description="Client Secret", env="CLIENT_SECRET")
    authority: str = Field(None, description="login authority", env="AUTHORITY")
    domain: str = Field(None, description="Domain name", env="DOMAIN")
    tenant_id: str = Field(None, description="Tenant Id", env="TENANT_ID")
    scopes: str = Field(None, description="Scopes", env="SCOPES")
    vault_name: str = Field(None, description="Global Vault Url", env="VAULT_NAME")
    vault_certificate_key: str = Field(
        None, description="Certificate name", env="VAULT_CERTIFICATE_KEY"
    )
    vault_secret_key: str = Field(
        None, description="Certificate name", env="VAULT_SECRET_KEY"
    )
    aad_issuers_list: List[str] = []

    @property
    def authorization_url(self):
        return f"{self.authority}/oauth2/v2.0/authorize"

    @property
    def token_url(self):
        return f"{self.authority}/oauth2/v2.0/token"

    @property
    def keys_url(self):
        return f"{self.authority}/discovery/v2.0/keys"

    @property
    def vault_url(self):
        return f"https://{self.vault_name}.vault.azure.net"

    def get_available_issuers(self):
        if self.aad_issuers_list is not None and len(self.aad_issuers_list) > 0:
            return self.aad_issuers_list

        issuers_list_url = (
            "https://login.microsoftonline.com/common/discovery/instance"
            + "?authorization_endpoint="
            + "https://login.microsoftonline.com/common/oauth2/v2.0/"
            + "authorize&api-version=1.1"
        )

        _issuers_list = requests.get(issuers_list_url).json()
        _metadatas = _issuers_list["metadata"]

        _authority_parser = urlparse(self.authority)
        _authority_domain = _authority_parser.hostname

        for _metadata in _metadatas:
            if _authority_domain in _metadata["preferred_network"]:
                for alias in _metadata["aliases"]:
                    self.aad_issuers_list.append(f"https://{alias}/{self.tenant_id}/")
                    self.aad_issuers_list.append(
                        f"https://{alias}/{self.tenant_id}/v2.0"
                    )

        return self.aad_issuers_list

    @property
    def audiences(self):
        """
        Returns the audiences that are valid for a client_id
        """
        return [
            f"https://{self.domain}/{self.client_id}",
            f"api://{self.client_id}",
            self.client_id,
        ]

    @property
    def scopes_list(self):

        _scopes = []

        if self.scopes is not None:
            if isinstance(self.scopes, typing.List):
                _scopes.extend(self.scopes)
            else:
                for _scp in str.split(self.scopes, " "):
                    _scopes.append(_scp)

        return _scopes

    # @property
    # def api_scopes_identifiers_root(self) -> List[str]:
    #     if self.api_scopes is None:
    #         return None

    #     scopes_identifiers = []
    #     for ls in self.api_scopes:
    #         scopes_identifiers.append(f"https://{self.domain}/{self.client_id}")

    #     return scopes_identifiers

    # @property
    # def api_scopes_identifiers(self) -> List[str]:
    #     if self.api_scopes is None:
    #         return None

    #     scopes_identifiers = []
    #     for ls in self.api_scopes:
    #         scopes_identifiers.append(f"https://{self.domain}/{self.client_id}/{ls}")

    #     return scopes_identifiers

    # @property
    # def api_scopes_identifiers_default(self) -> List[str]:
    #     if self.api_scopes is None:
    #         return None

    #     scopes_identifiers = []
    #     for ls in self.api_scopes:
    #         scopes_identifiers.append(
    #             f"https://{self.domain}/{self.client_id}/.default"
    #         )

    #     return scopes_identifiers

    # @property
    # def graph_scopes_identifiers_default(self) -> List[str]:
    #     return ["https://graph.microsoft.com/.default"]

    class Config:
        env_file = ".env"



import json

import pytest
from authlib.jose.rfc7519.claims import JWTClaims
from helpers import gen_payload

from aad import AadUser, AuthError, AzureAdSettings
from aad.aad_helpers import _decode_token, _get_user_from_claims, _validate_claims


def test_decode_token(private_key, public_key):
    """Test case for aad_helpers._decode_token

    Ensures we can decode a valid token
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    assert isinstance(claims, JWTClaims)


def test_decode_token_fail_if_no_key(private_key, public_key):
    """Test case for aad_helpers._decode_token

    Ensures that we have a correct error raised if key is invalid
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    access_token = token["access_token"]

    with pytest.raises(Exception) as auth_error:
        _decode_token(access_token)

    assert auth_error is not None
    assert isinstance(auth_error.value, AuthError)
    assert auth_error.value.code == "invalid_public_key"


def test_validate_claims_with_tenant_id(private_key, public_key):
    """Test case for aad_helpers._validate_claims

    Ensures we have valid claims, and that issuer is validated
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    _validate_claims(claims, issuers=options.tenant_id)

    assert True


def test_validate_claims(private_key, public_key):
    """Test case for aad_helpers._validate_claims

    Ensures we have valid claims, and that issuer is validated
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    _validate_claims(claims, issuers=claims.get("tid"))

    assert True


def test_validate_claims_multi_audiences(private_key, public_key):
    """Test case for aad_helpers._validate_claims

    Ensures we have valid claims, and that issuer is validated
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    aud = claims.get("aud")
    aud_list = [aud, "client_id"]
    _validate_claims(claims, issuers=claims.get("tid"), audiences=aud_list)

    assert True


def test_validate_claims_with_issuers_list(private_key, public_key):
    """Test case for aad_helpers._validate_claims

    Ensures we have valid claims, and that issuer is validated
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(
        options, private_key, iss="http://a_valid_issuer/a_valid_host_name"
    )
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    issuers = [
        "http://a_valid_issuer/a_valid_host_name",
        "http://a_valid_issuer/a_valid_host_name/v2.0",
    ]

    _validate_claims(claims, issuers=issuers)

    assert True


def test_validate_claims_should_fail(private_key, public_key):
    """Test case for aad_helpers._validate_claims

    Ensures we are rejecting claims is issuer is not valid
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(
        options, private_key, iss="http://not_the_same_issuer_than_token"
    )
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)

    with pytest.raises(Exception):
        _validate_claims(claims, issuer_tenant_id=options.tenant_id)


def test_get_user_from_claims(private_key, public_key):
    """Test case for aad_helpers._decode_token

    Ensures we have a valid user from claims
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    claims = _decode_token(token["access_token"], public_key=public_key)
    user = _get_user_from_claims(claims)

    assert isinstance(user, AadUser)
    assert user.username == f"John.Doe@{options.domain}"



import base64
import binascii
from typing import Dict

from azure.identity import DefaultAzureCredential
from azure.keyvault.certificates import CertificateClient
from azure.keyvault.secrets import SecretClient
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs12


class AadVault:
    def __init__(self, vault_url: str):
        self.vault_url = vault_url

    def get_certificate(self, vault_certificate_key: str):
        credential = DefaultAzureCredential()

        certificate_client = CertificateClient(self.vault_url, credential)
        secret_client = SecretClient(self.vault_url, credential)

        # Get the certificate secret that contains the private key
        certificate_secret = secret_client.get_secret(name=vault_certificate_key)

        # Get the certificate that contains the encoded thumbprint
        certificate = certificate_client.get_certificate(
            certificate_name=vault_certificate_key
        )

        # Get the hexadecimal thumbprint from X509 binary thumbprint value
        thumbprint = binascii.hexlify(certificate.properties.x509_thumbprint)

        certificate = {
            "private_key": certificate_secret.value,
            "thumbprint": thumbprint,
            "x_type": certificate_secret.properties._content_type,
        }

        return certificate

    def get_secret(self, vault_secret_key: str):
        credential = DefaultAzureCredential()
        secret_client = SecretClient(self.vault_url, credential)

        # Get the certificate secret that contains the private key
        secret = secret_client.get_secret(name=vault_secret_key)

        return secret.value

    def get_msal_client_credential(self, certificate: Dict):
        # get certificate type
        thumbprint = certificate["thumbprint"]
        private_key = certificate["private_key"]
        content_type = certificate.get("x_type", "x-pem-file")

        if "x-pem-file" in content_type:
            # Create the credential
            client_credential = {
                "thumbprint": thumbprint,
                "private_key": serialization.load_pem_private_key(
                    private_key.encode(), password=None
                ),
            }
        else:
            # Get the bytes from the base64 value
            cert_bytes = base64.b64decode(private_key)

            (
                private_key,
                public_certificate,
                additional_certificates,
            ) = pkcs12.load_key_and_certificates(data=cert_bytes, password=None)

            client_credential = {
                "thumbprint": thumbprint,
                "private_key": private_key,
            }

        return client_credential



import json
from collections import namedtuple

import pytest
from azure.identity._credentials.chained import ChainedTokenCredential
from helpers import gen_payload
from http_client import MinimalResponse

from aad import (
    AadClient,
    AuthError,
    AuthToken,
    ensure_user_from_token,
)


def test_aad_client_options_is_set():
    """Test case for AadAuthenticationClient ctor

    Ensures that options are set correctly
    """
    aad_test = AadClient()
    assert aad_test.options is not None


@pytest.mark.asyncio
async def test_aad_client_options_is_not_set(client_credential):
    """Test case for AadAuthenticationClient ctor

    Ensures that options not set will fail correctly
    """
    aad_client = AadClient(options={})

    with pytest.raises(Exception) as auth_error:
        await aad_client.build_msal_confidential_app(
            client_credential=client_credential.__dict__
        )

    assert auth_error is not None
    assert isinstance(auth_error.value, AuthError)
    assert auth_error.value.code == "options_client_id_missing"


@pytest.mark.asyncio
async def test_construct_flow(client_credential):
    """Test case for aad_client.build_auth_code_flow(scopes, redirect_uri)

    Ensures that flow is correctly constructed and return redirect_uri
    """

    # azure aad client
    aad_client = AadClient()

    scopes_identifiers = aad_client.options.api_scopes_identifiers
    redirect_uri = "https://localhost/oauth-redirect.html"

    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, redirect_uri, client_credential=client_credential.__dict__
    )

    assert flow is not None
    assert "redirect_uri" in flow
    assert flow["redirect_uri"] == redirect_uri
    assert "auth_uri" in flow
    assert "state" in flow
    assert "code_verifier" in flow
    assert "scope" in flow
    for scope in scopes_identifiers:
        assert scope in flow["scope"]


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow(client_credential):
    """Test case for aad_client.acquire_token_by_auth_code_flow using a user credential

    Ensures that user can acquire a valid token
    and get the session user filled with correct values
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert user is not None
    assert user.name == "John Doe"
    assert user.auth_token is not None
    assert user.auth_token.access_token is not None
    assert user.auth_token.client_info is not None
    assert aad_client.session is not None


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_through_general_method(
    client_credential,
):
    """Test case for aad_client.acquire_token_by_auth_code_flow using a user credential

    Ensures that user can acquire a valid token
    and get the session user filled with correct values
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client.acquire_user_token(
        scopes_identifiers,
        auth_response=auth_response,  # fake auth response from flow
        flow=flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert user is not None
    assert user.name == "John Doe"
    assert user.auth_token is not None
    assert user.auth_token.access_token is not None
    assert user.auth_token.client_info is not None
    assert aad_client.session is not None


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_with_roles(client_credential):
    """Test case for aad_client.acquire_token_by_auth_code_flow
    using a user credential and roles

    Ensures that roles are correctly retrieved from token
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options,
            client_credential.private_key,
            roles=["Admin", "User"],
            scp=scopes_identifiers,
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )
    assert user is not None
    assert user.auth_token is not None
    assert user.auth_token.client_info is not None
    assert user.auth_token.access_token is not None
    assert user.roles_id is not None
    assert len(user.roles_id) == 2
    assert "Admin" in user.roles_id
    assert "User" in user.roles_id


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_with_scopes(client_credential):
    """Test case for aad_client.acquire_token_by_auth_code_flow
    using a user credential and scopes

    Ensures that scopes are correctly retrieved from token
    """

    # azure aad client
    aad_client = AadClient()
    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert user.scopes is not None
    assert len(user.scopes) == len(scopes_identifiers) + 1
    for scope in aad_client.options.scopes:
        assert scope in user.scopes


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_with_scope(client_credential):
    """Test case for aad_client.acquire_token_by_auth_code_flow
    using a user credential and one scope

    Ensures that one scope is correctly retrieved from token (if only 1 scope)
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scope_identifier = aad_client.options.api_scopes_identifiers[0]

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scope_identifier, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scope_identifier
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scope_identifier,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert user.scopes is not None
    assert len(user.scopes) == 2
    assert aad_client.options.scopes[0] in user.scopes


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_with_scope_then_acquire_on_behalf_of(
    client_credential,
):
    """Test case for aad_client.acquire_token_by_auth_code_flow
    using a user credential and one scope

    Ensures that one scope is correctly retrieved from token (if only 1 scope)
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scope_identifier = aad_client.options.api_scopes_identifiers[0]

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scope_identifier, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scope_identifier
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scope_identifier,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    def mock_post_2(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=["User.Read"]
        )
        return MinimalResponse(status_code=200, text=payload)

    token2 = await aad_client._acquire_token_on_behalf_of(
        user.auth_token,
        ["User.Read"],
        post=mock_post_2,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert token2 is not None
    from aad.aad_helpers import _decode_token

    decoded_token = _decode_token(token2, public_key=client_credential.public_key)
    assert decoded_token is not None
    assert "scp" in decoded_token
    assert len(decoded_token["scp"]) > 0
    assert decoded_token["scp"][0] == "User.Read"


@pytest.mark.asyncio
async def test_acquire_token_by_auth_code_flow_with_wrong_public_key(client_credential):
    """Test case for aad_client.acquire_token using a user credential

    Ensures that decoding the token with an incorrect public key will fail
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key)
        return MinimalResponse(status_code=200, text=payload)

    public_key = client_credential.public_key.replace("DwIDAQAB", "DwIDAQAA")

    with pytest.raises(Exception) as auth_error:
        await aad_client._acquire_token_and_user_by_auth_code_flow(
            scopes_identifiers,
            auth_response,  # fake auth response from flow
            flow,  # correct flow from azure ad
            post=mock_post,
            public_key=public_key,
            client_credential=client_credential.__dict__,
        )

    assert auth_error is not None
    assert isinstance(auth_error.value, AuthError)
    assert auth_error.value.code == "bad_signature"


@pytest.mark.asyncio
async def test_acquire_token_on_behalf_of(client_credential, private_key):
    """Test case for aad_client.acquire_token using a service principal

    Ensures that service principal can acquire a valid token and
    get the session user filled with correct values
    """
    # azure aad client
    aad_client = AadClient()

    payload = gen_payload(aad_client.options, private_key)
    token = json.loads(payload)
    auth_token = AuthToken(token=token)

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key, True)
        return MinimalResponse(status_code=200, text=payload)

    token = await aad_client._acquire_token_on_behalf_of(
        auth_token,
        ["User.Read"],
        post=mock_post,
        public_key=client_credential.public_key,
        client_credential=client_credential.__dict__,
    )

    assert token is not None
    assert token.client_info is None
    assert token.access_token is not None


@pytest.mark.asyncio
async def test_acquire_token_on_behalf_of_through_general_method(
    client_credential, private_key
):
    """Test case for aad_client.acquire_token using a service principal

    Ensures that service principal can acquire a valid token and
    get the session user filled with correct values
    """
    # azure aad client
    aad_client = AadClient()

    payload = gen_payload(aad_client.options, private_key, scp="bjiu-123")
    token = json.loads(payload)
    auth_token = AuthToken(token=token)
    user = ensure_user_from_token(auth_token, public_key=client_credential.public_key)

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp="User.Read"
        )
        return MinimalResponse(status_code=200, text=payload)

    user = await aad_client.acquire_user_token(
        user=user,
        scopes=["User.Read"],
        post=mock_post,
        public_key=client_credential.public_key,
        client_credential=client_credential.__dict__,
    )

    assert user is not None
    assert user.auth_token is not None
    assert user.auth_token.access_token is not None


@pytest.mark.asyncio
async def test_acquire_token_for_managed_identity(
    client_credential, monkeypatch: pytest.MonkeyPatch
):

    # azure aad client
    aad_client = AadClient()

    def mock_get_token(self, scopes, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key, True)
        auth_token = json.loads(payload)
        AccessToken = namedtuple("AccessToken", ["token", "expires_on"])
        accessToken = AccessToken(token=auth_token["access_token"], expires_on=None)
        return accessToken

    monkeypatch.setattr(ChainedTokenCredential, "get_token", mock_get_token)

    token = await aad_client._acquire_token_for_managed_identity()

    assert token is not None
    assert token.client_info is None
    assert token.access_token is not None


@pytest.mark.asyncio
async def test_acquire_token_from_cache(client_credential):
    """Test case for aad_client.acquire_token using a user credential

    Ensures that user can acquire a valid token from AD,
    then acquiring this token from cache if valid
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key)
        return MinimalResponse(status_code=200, text=payload)

    await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    # 2nd call should it the cache
    # no need to set flow and auth_response since everything should happen from cache
    await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    # 3nd call should it the cache
    # no need to set flow and auth_response since everything should happen from cache
    user_3 = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    assert user_3 is not None
    assert user_3.name == "John Doe"
    assert user_3.auth_token is not None
    assert user_3.auth_token.access_token is not None
    assert aad_client.session is not None


@pytest.mark.asyncio
async def test_acquire_token_from_flow_works_without_session_set(client_credential):

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key)
        return MinimalResponse(status_code=200, text=payload)

    #
    # flow is not explicitely set, so should be in a session object
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )
    assert user is not None
    assert user.name == "John Doe"
    assert user.auth_token is not None
    assert user.auth_token.access_token is not None
    assert user.auth_token.client_info is not None
    assert aad_client.session is not None


@pytest.mark.asyncio
async def test_build_msal_public():
    # azure aad client
    aad_client = AadClient()

    msal_public = await aad_client.build_msal_public_app()

    assert msal_public is not None


@pytest.mark.asyncio
async def test_get_client(client_credential):

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    user = await aad_client.get_user(False)

    assert user is not None
    assert user.name == "John Doe"
    assert user.auth_token is not None
    assert user.auth_token.access_token is not None
    assert aad_client.session is not None


@pytest.mark.asyncio
async def test_remove_account(client_credential):

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    aad_id = aad_client.session.get("aad_id")
    assert aad_id is not None
    session_cache = await aad_client.cache_manager.get(aad_id)
    assert session_cache is not None

    await aad_client.remove_account(
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    aad_id = aad_client.session.get("aad_id")
    assert aad_id is None
    session_cache = await aad_client.cache_manager.get(aad_id)
    assert isinstance(session_cache, dict)
    assert len(list(session_cache)) == 0



from fastapi import APIRouter
from starlette.requests import Request

router = APIRouter(
    prefix="/engines",
    tags=["engines"],
    responses={404: {"description": "Not found"}},
)

fake_items_db = {"plumbus": {"name": "Plumbus"}, "gun": {"name": "Portal Gun"}}

company = {
    "company_id": "000ABC",
    "company_name": "One Two Three",
    "id": "601BBA98-677A-49E8-B677-6905C62D521C",
    "webapis_oids": [
        {
            "id": "36AA308B-5B0F-447A-BABC-40330B004F16",
            "app_name": "OneTwoThreeWebAppService",
            "roles_id": ["data-provider", "data-browser"],
        }
    ],
}


# Always check scopes for API auth
@router.get("")
async def get_engines(request: Request):
    return company



import pytest
from helpers import gen_payload
from http_client import MinimalResponse
from requests.models import CaseInsensitiveDict
from starlette.testclient import TestClient

from aad import AadClient, AadBearerBackend, AadUser


def test_aad_server_options_is_set():
    """Test case for AadBearerMiddleware(app)

    Ensures that options can be set automatically
    """
    aad_server_options = AadBearerBackend()
    assert aad_server_options.options is not None
    assert aad_server_options.options.client_id is not None


@pytest.mark.asyncio
async def test_is_not_authorized_if_no_bearer_token(client: TestClient):
    """Test case for token=Depends(oauth2_scheme())

    Ensures we can't access an api with authorization set
    """

    response = await client.get("/isauth")
    assert response.status_code == 403
    response_details = response.json()
    assert response_details["detail"] == "Not authenticated"


@pytest.mark.asyncio
async def test_user_is_authorized_from_authenticated_request(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Ensures that we can send a request and get a response
    if authenticated with a user credential
    """
    # azure aad client
    aad_client = AadClient()

    # get a confidential app
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key)
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"

    response = await client.get("/isauth", headers=headers)
    assert response.status_code == 200
    assert response.content is not None

    user = AadUser.create_user(**response.json())
    assert user.is_authenticated is True


@pytest.mark.asyncio
async def test_service_principal_is_authorized_from_authenticated_request(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Ensures that we can send a request and get a response if
    authenticated with a service principal
    """

    # azure aad client
    aad_client = AadClient()

    # build confidential app
    scopes_identifiers = aad_client.options.api_scopes_identifiers_default

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options,
            client_credential.private_key,
            True,
            scp=scopes_identifiers,
        )
        return MinimalResponse(status_code=200, text=payload)

    token = await aad_client._acquire_token_for_client(
        scopes_identifiers,
        post=mock_post,
        public_key=client_credential.public_key,
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {token.access_token}"
    response = await client.get("/isauth", headers=headers)

    assert response.status_code == 200
    assert response.content is not None

    user_from_api = AadUser.create_user(**response.json())
    assert user_from_api.is_authenticated is True


@pytest.mark.asyncio
async def test_service_principal_is_unauthorized_if_scopes_are_needed(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test case for @authorize("user_impersonation")

    Ensures that a service principal can't access an api with scopes
    """

    # azure aad client
    aad_client = AadClient()

    # build confidential app
    scopes_identifiers = aad_client.options.api_scopes_identifiers_default

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(aad_client.options, client_credential.private_key, True)
        return MinimalResponse(status_code=200, text=payload)

    token = await aad_client._acquire_token_for_client(
        scopes_identifiers,
        post=mock_post,
        public_key=client_credential.public_key,
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {token.access_token}"
    response = await client.get("/isauth_impersonation", headers=headers)
    assert response.status_code == 403
    detail_response = response.json()
    assert detail_response["detail"] == "Forbidden"


@pytest.mark.asyncio
async def test_user_is_authorized_from_authenticated_request_with_scopes(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test case for @authorize("user_impersonation")

    Ensures that we can send a user token with scopes
    and get a response if scopes are matching
    """

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"
    response = await client.get("/isauth_impersonation", headers=headers)
    assert response.status_code == 200
    assert response.content is not None

    user = AadUser.create_user(**response.json())
    assert user.is_authenticated is True
    assert len(user.scopes) == len(aad_client.options.scopes) + 1
    assert "authenticated" in user.scopes
    for scope in aad_client.options.scopes:
        assert scope in user.scopes


@pytest.mark.asyncio
async def test_user_is_unauthorized_from_authenticated_request_if_scopes_mismatch(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test case for @authorize("user_impersonation")

    Ensures that a user token with bad scopes is
    not authorized to acces an api with scopes specified
    """

    # azure aad client
    aad_client = AadClient()

    # build code flow
    # fake scopes that should fail
    scopes_identifiers = "other_scope_not_good"

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"

    response = await client.get("/isauth_impersonation", headers=headers)

    assert response.content is not None
    assert response.status_code == 403
    detail_response = response.json()
    assert detail_response["detail"] == "Forbidden"


@pytest.mark.asyncio
async def test_user_is_authorized_from_authenticated_request_with_roles(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test for @authorize("user_impersonation", ["Admin", "Contributor"])

    Ensures that a user token with scopes and roles is
    authorized to acces an api with scopes / roles specified
    """

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # fake roles
    roles = ["Admin", "Contributor"]

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options,
            client_credential.private_key,
            scp=scopes_identifiers,
            roles=roles,
        )
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"

    response = await client.get("/isauth_impersonation_roles", headers=headers)

    assert response.status_code == 200
    assert response.content is not None

    user = AadUser.create_user(**response.json())
    assert user.is_authenticated is True
    assert len(user.roles_id) == len(roles)
    for role in roles:
        assert role in user.roles_id


@pytest.mark.asyncio
async def test_user_is_unauthorized_from_authenticated_request_if_roles_mismatch(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test case for @authorize("user_impersonation", ["Admin", "Contributor"])

    Ensures that a user token with bad roles
    is not authorized to acces an api with scopes / roles specified
    """

    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # fake roles
    # we miss the "Admin" role here, and the API needs it
    roles = ["Contributor"]

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options,
            client_credential.private_key,
            scp=scopes_identifiers,
            roles=roles,
        )
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"

    response = await client.get("/isauth_impersonation_roles", headers=headers)

    assert response.status_code == 403
    detail_response = response.json()
    assert detail_response["detail"] == "Unauthorized role"


@pytest.mark.asyncio
async def test_user_is_unauthorized_from_authenticated_request_if_no_roles(
    client: TestClient, client_credential
):
    """Test case for token=Depends(oauth2_scheme())

    Test case for @authorize("user_impersonation", ["Admin", "Contributor"])

    Ensures that a user token without roles
    is not authorized to acces an api with scopes / roles specified
    """
    # azure aad client
    aad_client = AadClient()

    # build code flow
    scopes_identifiers = aad_client.options.api_scopes_identifiers

    # constructing flow to have the auth uri
    flow = await aad_client.build_auth_code_flow(
        scopes_identifiers, client_credential=client_credential.__dict__
    )

    # build fake response from azure ad with good state / fake code
    auth_response = {"state": flow["state"], "code": "012"}

    def mock_post(url, headers=None, *args, **kwargs):
        payload = gen_payload(
            aad_client.options, client_credential.private_key, scp=scopes_identifiers
        )
        return MinimalResponse(status_code=200, text=payload)

    # this method is alrady tested
    user = await aad_client._acquire_token_and_user_by_auth_code_flow(
        scopes_identifiers,
        auth_response,  # fake auth response from flow
        flow,  # correct flow from azure ad
        post=mock_post,  # fake response
        public_key=client_credential.public_key,  # public key to decrypt fake response
        client_credential=client_credential.__dict__,
    )

    headers = CaseInsensitiveDict()
    headers["Accept"] = "application/json"
    headers["Authorization"] = f"Bearer {user.auth_token.access_token}"

    response = await client.get("/isauth_impersonation_roles", headers=headers)

    assert response.status_code == 403
    detail_response = response.json()
    assert detail_response["detail"] == "Unauthorized role"



class AuthError(Exception):
    code: str

    def __init__(
        self,
        code: str = None,
        description: str = None,
        status_code=401,
        exception: Exception = None,
    ):

        if exception is not None and code is None:
            if hasattr(exception, "error"):
                code = exception.error
            elif hasattr(exception, "code"):
                code = exception.code

        self.code = code

        if exception is not None and description is None:
            if hasattr(exception, "message"):
                description = exception.message
            elif hasattr(exception, "description"):
                description = exception.description
            elif hasattr(exception, "args") and len(exception.args) >= 1:
                description = exception.args[0]

        self.description = description
        super().__init__(description)

        self.status_code = status_code



from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    UnauthenticatedUser,
)

from .aad_client import AadClient


class AadSessionBackend(AuthenticationBackend):
    async def authenticate(self, request):
        """Authenticate a request.
        If authentication is successful, defining a user instance
        """
        try:

            if not request.session or not request.session.get("aad_id"):
                return AuthCredentials(None), UnauthenticatedUser()

            aad_client = AadClient(session=request.session)

            # Do not validate signature, since we may have here a
            # microsoft graph token, that we can't validate
            # but it's fine since we are not on the web api side
            user = await aad_client.get_user(False)

            if user is None:
                return AuthCredentials(None), UnauthenticatedUser()

            return AuthCredentials(user.scopes), user

        except Exception:
            return AuthCredentials(None), UnauthenticatedUser()



import time
from asyncio import Lock, create_task, sleep
from typing import Dict, Optional


class CacheManager:
    cache_db: Dict[any, any] = dict()
    _lock = Lock()

    @classmethod
    async def set(cls, key: str, value: Dict[str, any]) -> None:
        async with cls._lock:

            if "__expires_at__" not in value:
                expires_in = 86400
                expires_at = time.time() + expires_in
                value["__expires_at__"] = expires_at

            cls.cache_db.update({key: value})

    @classmethod
    async def get(cls, key: str) -> Optional[Dict[str, any]]:
        async with cls._lock:
            return cls.cache_db.get(key, {})

    @classmethod
    async def remove(cls, key: str) -> None:
        async with cls._lock:
            cls.cache_db.pop(key, None)

    @classmethod
    async def clear(cls) -> int:
        async with cls._lock:
            cls.cache_db.clear()

    @classmethod
    async def check_expiracy(cls) -> None:
        # copy list to prevent changing dict size during iteration
        for key in list(CacheManager.cache_db.keys()):
            value = await cls.get(key)
            if value is not None and "__expires_at__" in value:
                expires_at = int(value["__expires_at__"])
                if expires_at < int(time.time()):
                    value = await cls.get(key)
                    if value is not None:
                        del CacheManager.cache_db[key]

    @classmethod
    async def for_ever_check(cls, interval: int = 3600) -> None:
        """sets a never ending task for checking cache expired tasks"""
        while True:
            # wait until next run
            await sleep(interval)
            await cls.check_expiracy()

    @classmethod
    def start_expiracy_daemon(cls, interval: int = 3600) -> None:
        """starts the daemon task for checking cache expired tasks"""

        create_task(cls.for_ever_check(interval))



import json
import os
import pathlib
from os import environ

import msal
import requests
from aad_fastapi import AadClient, AuthError, AuthToken, ensure_user_from_token
from dotenv import load_dotenv
from flask import Flask, redirect, render_template, request, session, url_for
from flask_session import Session  # https://pythonhosted.org/Flask-Session
from werkzeug.middleware.proxy_fix import ProxyFix

app = Flask(__name__)
app.config["SESSION_TYPE"] = "filesystem"
Session(app)

# This section is needed for url_for("foo", _external=True) to automatically
# generate http scheme when this sample is running on localhost,
# and to generate https scheme when it is deployed behind reversed proxy. See also
# https://flask.palletsprojects.com/en/1.0.x/deploying/wsgi-standalone/#proxy-setups
app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1, x_host=1)

client_id = os.environ["CLIENT_ID"]
domain = os.environ["DOMAIN"]
api_url = environ.get("API_URL")

dir = pathlib.Path(__file__).parent.parent.absolute()
localenv = os.path.join(dir, "local.env")
if os.path.exists(localenv):
    load_dotenv(localenv, override=True)


@app.route("/", methods=["GET", "POST"])
async def index():
    try:

        aad_client = AadClient(session=session)

        redirect_url = url_for("authorized", _external=True)

        flow = await aad_client.build_auth_code_flow(
            aad_client.options.scopes_list, redirect_url
        )

        if request.method == "POST":
            token = request.form["txt_access_token"]
            auth_token = AuthToken(token)
            if token is not None and token != "":
                user = ensure_user_from_token(auth_token, False)
                session["access_token"] = token
                session["claims"] = user.claims
                session["user"] = user.username
        else:
            session["data"] = None

        user = await aad_client.get_user()

        username = user.name if user is not None else ""
        claims = json.dumps(user.claims) if user is not None else None
        message = session.get("message", "")
        access_token = session.get("access_token", "")
        data = json.dumps(session["data"]) if "data" in session else ""

        return render_template(
            "index.html",
            username=username,
            auth_url=flow["auth_uri"],
            version=msal.__version__,
            claims=claims,
            message=message,
            access_token=access_token,
            data=data,
        )

    except AuthError as aex:
        return render_template(
            "auth_error.html",
            result={"error": aex.code, "error_description": aex.description},
        )
    except Exception as ex:
        return render_template(
            "auth_error.html",
            result={"error": "Unknwon error", "error_description": ex.args[0]},
        )


@app.route("/userlogin")
async def userlogin():
    aad_client = AadClient(session=session)
    redirect_url = url_for("authorized", _external=True)

    flow = await aad_client.build_auth_code_flow(
        aad_client.options.scopes_list, redirect_url
    )

    auth_url = flow["auth_uri"]
    return redirect(auth_url)


@app.route(
    "/auth/oauth2-redirect"
)  # Its absolute URL must match your app's redirect_uri set in AAD
async def authorized():
    try:
        aad_client = AadClient(session=session)
        user = await aad_client._acquire_token_and_user_by_auth_code_flow(
            aad_client.options.scopes_list, request.args
        )

        session["access_token"] = user.auth_token.access_token

    except AuthError as ae:
        print(ae)
        return render_template(
            "auth_error.html",
            result={"error": ae.code, "error_description": ae.description},
        )
    except ValueError as ve:  # Usually caused by CSRF
        print(ve)
        pass  # Simply ignore them
    except Exception as ex:
        print(ex)
        return render_template(
            "auth_error.html",
            result={"error": "Unknown", "error_description": ex.args[0]},
        )

    return redirect(url_for("index"))


@app.route("/logout")
def logout():
    session.clear()  # Wipe out user and its token cache from session
    aad_client = AadClient(session=session)
    return redirect(
        aad_client.get_logout_uri(url_for("index", _external=True))
    )  # Also logout from your tenant's web session


@app.route("/apicall")
async def apicall():

    try:
        aad_client = AadClient(session=session)
        valid_user = await aad_client.acquire_user_token()

    except AuthError:
        return redirect(url_for("userlogin"))

    engines = requests.get(f"{api_url}/engines", auth=valid_user.auth_token).json()

    return render_template("display.html", result=engines)


if __name__ == "__main__":
    app.run()



from typing import Dict

from aiohttp.helpers import BasicAuth
from fastapi.security.utils import get_authorization_scheme_param
from requests.auth import AuthBase

from aad.aad_autherror import AuthError


class AuthToken(AuthBase, BasicAuth):
    def __init__(
        self,
        access_token: str = None,
        id_token: str = None,
        refresh_token: str = None,
        client_info: str = None,
        token: Dict[str, str] = None,
    ):

        self.access_token = access_token
        self.id_token = id_token
        self.client_info = client_info
        self.refresh_token = refresh_token

        if token:
            if "access_token" in token:
                self.access_token = token["access_token"]
            if "id_token" in token:
                self.id_token = token["id_token"]
            if "client_info" in token:
                self.client_info = token["client_info"]
            if "refresh_token" in token:
                self.refresh_token = token["refresh_token"]

        if self.access_token is None:
            raise AuthError(
                "access_token_empty", "No access token in the BearerAuth instance"
            )

        if self.access_token.lower().startswith("bearer"):
            _, self.access_token = get_authorization_scheme_param(self.access_token)

    # Used by BasicAuth
    def __new__(
        cls,
        login: str = "",
        password: str = "",
        encoding: str = "latin1",
        *args,
        **kwargs,
    ) -> BasicAuth:
        # even if not used, mandatory to be sure instance is not returning False
        login = login if login is not None and login != "" else "unused"
        return super().__new__(cls, login, password, encoding)

    # Used by BasicAuth
    def encode(self) -> str:
        return f"Bearer {self.access_token}"

    # User by AuthBase
    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.access_token
        return r



from aad import oauth2_scheme
from aad.aad_options import AzureAdSettings


def test_oauth2_scheme_options_is_set():
    """Test case for AadBearerMiddleware(app)

    Ensures that options can be set automatically
    """
    oauth2 = oauth2_scheme()

    assert oauth2 is not None
    assert oauth2.scheme_name == "OAuth2"
    assert oauth2.model is not None
    assert oauth2.model.flows is not None
    assert oauth2.model.flows.authorizationCode is not None


def test_oauth2_scheme_custom_options_is_set():
    """Test case for AadBearerMiddleware(app)

    Ensures that options can be set automatically
    """
    options = AzureAdSettings()
    options.authority = "https://login.microsoftonline.com/tenant_id"
    options.scopes = "scope1 scope2"
    options.domain = "contoso.com"
    options.client_id = "A0A1A2-B0B1B2"

    oauth2 = oauth2_scheme(options=options)

    assert oauth2 is not None
    assert oauth2.scheme_name == "OAuth2"
    assert oauth2.model is not None
    assert oauth2.model.flows is not None
    assert oauth2.model.flows.authorizationCode is not None
    assert (
        oauth2.model.flows.authorizationCode.authorizationUrl
        == "https://login.microsoftonline.com/tenant_id/oauth2/v2.0/authorize"
    )
    assert len(oauth2.model.flows.authorizationCode.scopes) == 2
    assert (
        list(oauth2.model.flows.authorizationCode.scopes)[0]
        == "https://contoso.com/A0A1A2-B0B1B2/scope1"
    )
    assert (
        list(oauth2.model.flows.authorizationCode.scopes)[1]
        == "https://contoso.com/A0A1A2-B0B1B2/scope2"
    )



import typing

from authlib.jose import errors as jwt_errors
from authlib.jose import jwk, jwt
from authlib.jose.rfc7519.claims import JWTClaims
from authlib.jose.util import extract_header
from msal.oauth2cli.oidc import decode_id_token

from .aad_autherror import AuthError
from .aad_discover_keys import AadDiscoverKey
from .aad_options import AzureAdSettings
from .aad_token import AuthToken
from .aad_user import AadUser


def _decode_token_without_validating_signature(
    auth_token: typing.Union[AuthToken, str]
) -> JWTClaims:

    auth_token = AuthToken(auth_token) if isinstance(auth_token, str) else auth_token

    token = auth_token.access_token

    decoded = decode_id_token(token)
    # this Guid is the Microsoft Graph audience identifier
    c = JWTClaims(decoded, None)
    return c


def _decode_token(
    auth_token: typing.Union[AuthToken, str], keys_url=None, **kwargs
) -> JWTClaims:
    """Decode an access token using a public key or a key to retrieve the public key"""

    try:

        public_key = kwargs.pop("public_key", None)

        if keys_url is None and public_key is None:
            raise AuthError(
                "invalid_public_key",
                "public_key or key_url must be specified to be able to get the "
                "public key to decode the token",
            )

        jwkey = None

        auth_token = (
            AuthToken(auth_token) if isinstance(auth_token, str) else auth_token
        )

        # if id token exists, gets it
        # in that cicumstances, we are sure we can validate the token
        token = (
            auth_token.access_token
            if auth_token.id_token is None
            else auth_token.id_token
        )

        if public_key is not None:
            jwkey = public_key
        else:
            # get keys
            token_header = token.split(".")[0].encode()
            jwks = AadDiscoverKey.get_discovery_key_json(keys_url)
            unverified_header = extract_header(token_header, jwt_errors.DecodeError)

            # get signing keys, to be able to decode token
            for key in jwks["keys"]:
                if key["kid"] == unverified_header["kid"]:
                    jwkey = jwk.loads(key)

        claims = jwt.decode(token, jwkey)

        return claims

    # If the token is coming from Graph, we can't validate the signature
    # https://docs.microsoft.com/en-us/azure/active-directory/develop/access-tokens#validating-tokens
    except jwt_errors.BadSignatureError as bse:
        raise AuthError(exception=bse)
    except AuthError as aex:
        raise aex
    except Exception as ex:
        raise AuthError(exception=ex)


def _validate_claims(
    claims: JWTClaims,
    audiences: typing.Union[str, typing.List[str]] = None,
    issuers: typing.Union[str, typing.List[str]] = None,
):
    """Validate a claims."""

    claims_options = {}

    if audiences is not None and isinstance(audiences, str):
        aud_list = [audiences]
    else:
        aud_list = audiences

    if aud_list is not None and isinstance(aud_list, typing.List):
        claims_options["aud"] = {
            "essential": True,
            "values": aud_list,
        }

    # You can get a list of issuers for the various Azure AD deployments
    # (global & sovereign) from the following endpoint :
    # https://login.microsoftonline.com/common/discovery/instance
    # ?authorization_endpoint=
    # https://login.microsoftonline.com/common/oauth2/v2.0/authorize&api-version=1.1

    if issuers is not None and isinstance(issuers, str):
        issuers = [
            f"https://login.microsoftonline.com/{issuers}/",
            f"https://login.microsoftonline.com/{issuers}/v2.0",
            f"https://login.windows.net/{issuers}/",
            f"https://login.microsoft.com/{issuers}/",
            f"https://sts.windows.net/{issuers}/",
        ]

    if issuers is not None and isinstance(issuers, typing.List):
        claims_options["iss"] = {"essential": True, "values": issuers}

    claims.options = claims_options

    claims.validate()


def ensure_user_from_token(
    auth_token: AuthToken,
    validate: bool = True,
    options: AzureAdSettings = None,
    env_path: str = None,
    **kwargs,
):

    if options is None:
        options = AzureAdSettings(_env_file=env_path)

    if validate:
        claims = _decode_token(
            auth_token,
            keys_url=options.keys_url,
            **kwargs,
        )
    else:
        claims = _decode_token_without_validating_signature(auth_token)

    user = _get_user_from_claims(claims)
    user.auth_token = auth_token
    return user


def _get_user_from_claims(claims: JWTClaims) -> AadUser:
    """Get an AadUser from claims"""

    # get scopes and add the "authenticated"
    scopes = ["authenticated"]

    # scopes can be a list or an array
    if claims.get("scp") is not None:
        scp = claims.get("scp")
        if isinstance(scp, typing.List):
            scopes.extend(scp)
        else:
            for _scp in str.split(scp, " "):
                scopes.append(_scp)

    username_key = "preferred_username"
    name_key = "name"
    is_sp = False

    if username_key not in claims:
        username_key = "unique_name"

    if "name" not in claims:
        appid = "appid" if "appid" in claims else "azp"
        is_sp = True
        name_key = appid
        username_key = appid

    user = AadUser(claims[username_key])
    user.id = claims.get("oid", None)

    if user.id is None:
        raise AuthError("oid_is_missing" "Property oid is mandatory")

    user.name = claims[name_key]
    user.email = claims[username_key] if not is_sp else ""
    user.groups = claims.get("groups", None)
    user.roles_id = claims.get("roles", None)
    user.scopes = scopes
    user.claims = claims
    user.is_interactive = not is_sp
    return user



from .aad_client import AadClient
from .aad_decorators import authorize, oauth2_scheme
from .aad_autherror import AuthError
from .aad_bearer_backend import AadBearerBackend
from .aad_cache_manager import CacheManager
from .aad_discover_keys import AadDiscoverKey
from .aad_helpers import ensure_user_from_token
from .aad_options import AzureAdSettings
from .aad_session_backend import AadSessionBackend
from .aad_token import AuthToken
from .aad_user import AadUser

__all__ = [
    AzureAdSettings,
    AadBearerBackend,
    AadSessionBackend,
    authorize,
    oauth2_scheme,
    ensure_user_from_token,
    AadClient,
    AadUser,
    AuthError,
    AuthToken,
    AadDiscoverKey,
    CacheManager,
]



from typing import List, Optional

from authlib.jose.rfc7519.claims import JWTClaims
from starlette.authentication import SimpleUser

from .aad_token import AuthToken


class AadUser(SimpleUser):
    @staticmethod
    def create_user(**kwargs):

        user = AadUser(kwargs["username"])

        for key, value in kwargs.items():
            user.__dict__[key] = value

        return user

    id: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    username: str = None
    auth_token: Optional[AuthToken] = None
    roles_id: Optional[List[str]] = None
    groups: Optional[List[str]] = None
    scopes: Optional[List[str]] = None
    claims: Optional[JWTClaims] = None
    company: Optional[any] = None
    is_interactive: bool = True



import time

import pytest

from aad import CacheManager


@pytest.mark.asyncio
async def test_aad_cache_manager_set():

    await CacheManager.clear()
    key = "a"
    value = {"key": "value"}
    await CacheManager.set(key, value)

    get_value = await CacheManager.get(key)

    assert get_value.get("key") == "value"


@pytest.mark.asyncio
async def test_aad_cache_manager_expiracy():

    await CacheManager.clear()
    key = "a"
    value = {"key": "value"}
    await CacheManager.set(key, value)

    get_value = await CacheManager.get(key)

    assert get_value.get("__expires_at__") is not None


@pytest.mark.asyncio
async def test_aad_cache_manager_remove():

    await CacheManager.clear()
    key = "a"
    value = {"key": "value"}
    await CacheManager.set(key, value)

    assert len(CacheManager.cache_db) == 1
    await CacheManager.remove(key)
    assert len(CacheManager.cache_db) == 0


@pytest.mark.asyncio
async def test_aad_cache_manager_clear():

    await CacheManager.clear()
    key = "a"
    value = {"key": "value"}
    await CacheManager.set(key, value)
    key = "b"
    await CacheManager.set(key, value)

    assert len(CacheManager.cache_db) == 2
    await CacheManager.clear()
    assert len(CacheManager.cache_db) == 0


@pytest.mark.asyncio
async def test_aad_cache_manager_check_expiracy():

    await CacheManager.clear()
    key = "a"
    value = {"key": "value"}
    await CacheManager.set(key, value)
    value = await CacheManager.get(key)
    expires = int(value["__expires_at__"])
    assert expires > int(time.time())
    expires = int(time.time()) - 1000
    assert expires < int(time.time())
    # override expires
    value["__expires_at__"] = expires
    await CacheManager.set(key, value)
    await CacheManager.check_expiracy()
    assert len(CacheManager.cache_db) == 0



import json

import pytest
import requests
from helpers import gen_payload
from requests.models import Response

from aad import AadDiscoverKey, AuthError, AuthToken, AzureAdSettings


def test_discover_keys(monkeypatch: pytest.MonkeyPatch):

    keys = {
        "keys": [
            {
                "kty": "RSA",
                "use": "sig",
                "kid": "nOo3ZDrODXEK1jKWhXslHR_KXEg",
                "x5t": "nOo3ZDrODXEK1jKWhXslHR_KXEg",
                "n": "oaLLT9hkcSj2tGfZsjbu7Xz1Krs0",
                "e": "AQAB",
                "x5c": ["MIIDBTCCAe"],
                "issuer": "https://login.microsoftonline.com/aaaaaaa-/v2.0",
            }
        ]
    }
    data = json.dumps(keys)
    response = Response()
    response.status_code = 200
    response._content = data.encode()
    response.reason = "OK"
    monkeypatch.setattr(requests, "get", lambda url: response)

    keys = AadDiscoverKey.get_discovery_key_json("https://keys_url")

    assert keys is not None


def test_aad_error_status():
    err = AuthError(code="CODE", description="DESCRIPTION")

    assert err is not None
    assert err.status_code == 401

    err = AuthError(code="CODE", description="DESCRIPTION", status_code=403)

    assert err is not None
    assert err.status_code == 403


def test_aad_error_with_exception():
    class MyException(Exception):
        def __init__(self, error, message):
            self.error = error
            self.message = message

    err = AuthError(exception=MyException("ERROR", "MESSAGE"))

    assert err is not None
    assert err.code == "ERROR"
    assert err.description == "MESSAGE"

    class MyException2(Exception):
        def __init__(self, code, description):
            self.code = code
            self.description = description

    err = AuthError(exception=MyException("ERROR", "DESCRIPTION"))

    assert err is not None
    assert err.code == "ERROR"
    assert err.description == "DESCRIPTION"

    err = AuthError(exception=Exception("ERROR_DESCRIPTION"))

    assert err is not None
    assert err.code is None
    assert err.description == "ERROR_DESCRIPTION"


def test_auth_token_from_token(private_key):
    """Test case for auth_token

    Ensures we can decode a valid token
    """
    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    # add random values for id_token and refresh_token that
    # are not generated by the helper
    token["id_token"] = "ejy....id_token"
    token["refresh_token"] = "ejy....refresh_token"

    auth_token = AuthToken(token=token)

    assert auth_token.access_token is not None
    assert auth_token.id_token is not None
    assert auth_token.refresh_token is not None
    assert auth_token.client_info is not None
    assert auth_token.access_token == token["access_token"]
    assert auth_token.id_token == token["id_token"]
    assert auth_token.refresh_token == token["refresh_token"]
    assert auth_token.client_info == token["client_info"]


def test_auth_token_from_args(private_key):
    """Test case for auth_token

    Ensures we can decode a valid token
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)
    # add random values for id_token and refresh_token that
    # are not generated by the helper
    token["id_token"] = "ejy....id_token"
    token["refresh_token"] = "ejy....refresh_token"

    auth_token = AuthToken(
        token["access_token"],
        token["id_token"],
        token["refresh_token"],
        token["client_info"],
    )

    assert auth_token.access_token is not None
    assert auth_token.id_token is not None
    assert auth_token.refresh_token is not None
    assert auth_token.client_info is not None
    assert auth_token.access_token == token["access_token"]
    assert auth_token.id_token == token["id_token"]
    assert auth_token.refresh_token == token["refresh_token"]
    assert auth_token.client_info == token["client_info"]


def test_auth_token_fail_if_no_access_token():
    """Test case for AuthToken

    Ensures we can decode a valid token
    """

    # generate a valid payload
    with pytest.raises(Exception) as auth_error:
        AuthToken(id_token="ejy...")

    assert auth_error is not None
    assert isinstance(auth_error.value, AuthError)
    assert auth_error.value.code == "access_token_empty"


def test_auth_token_access_token_overrided_by_token(private_key):
    """Test case for AuthToken

    Ensures we can decode a valid token
    """

    # generate a valid payload
    options = AzureAdSettings()
    payload = gen_payload(options, private_key)
    token = json.loads(payload)

    auth_token = AuthToken(access_token="aaaaa", token=token)

    assert auth_token.access_token is not None
    assert auth_token.access_token == token["access_token"]



import os
import pathlib
from os import environ
from typing import cast
from dotenv import load_dotenv
from fastapi import Depends, FastAPI
from routers import engines
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import Request

from aad_fastapi import (
    AadBearerBackend,
    AadUser,
    authorize,
    oauth2_scheme,
    AzureAdSettings,
)

dir = pathlib.Path(__file__).parent.parent.absolute()
localenv = os.path.join(dir, "local.env")
if os.path.exists(localenv):
    load_dotenv(localenv, override=True)

# App Registration settings for protecting all the APIs.
api_options = AzureAdSettings()
api_options.client_id = environ.get("API_CLIENT_ID")
api_options.domain = environ.get("DOMAIN")
api_options.scopes = environ.get("SCOPES")

# App Registration setting for authentication SWAGGER WEB UI AUTHENTICATION.
web_ui_client_id = environ.get("CLIENT_ID")  # Client ID
web_ui_scopes = environ.get("SCOPES")  # Client ID

# pre fill client id
swagger_ui_init_oauth = {
    "usePkceWithAuthorizationCodeGrant": "true",
    "clientId": web_ui_client_id,
    "appName": "B-ID",
    "scopes": web_ui_scopes,
}

# Create a FasAPI instance
app = FastAPI(swagger_ui_init_oauth=swagger_ui_init_oauth)

# Add the bearer middleware, protected with Api App Registration
app.add_middleware(AuthenticationMiddleware, backend=AadBearerBackend(api_options))

# These routers needs an authentication for all its routes using Web App Registration
app.include_router(
    engines.router, dependencies=[Depends(oauth2_scheme(options=api_options))]
)


@app.get("/")
async def hello_world():
    try:
        return {"hello": "world"}
    except Exception as ex:
        return ex


@app.get("/user")
async def user(request: Request, token=Depends(oauth2_scheme(options=api_options))):

    try:
        return request.user
    except Exception as ex:
        return ex


@app.get("/user_with_scope")
@authorize("user_impersonation")
async def user_with_scope(
    request: Request, token=Depends(oauth2_scheme(options=api_options))
):

    user = cast(AadUser, request.user)

    try:
        return user
    except Exception as ex:
        return ex


@app.get("/user_with_scope_and_roles")
@authorize("user_impersonation", "security-administrator")
async def user_with_scope_and_roles(
    request: Request, token=Depends(oauth2_scheme(options=api_options))
):

    user = cast(AadUser, request.user)

    try:
        return user
    except Exception as ex:
        return ex



import json
from collections import namedtuple

import pytest
from helpers import gen_payload
from starlette.authentication import AuthCredentials, UnauthenticatedUser

from aad import AadSessionBackend, AadUser, AzureAdSettings
from aad.aad_cache_manager import CacheManager


@pytest.mark.asyncio
async def test_aad_session_backend_no_session():

    aad_session_backend = AadSessionBackend()
    request = namedtuple("request", ["session"])
    request.session = None

    cred, user = await aad_session_backend.authenticate(request=request)

    assert isinstance(cred, AuthCredentials)
    assert isinstance(user, UnauthenticatedUser)


@pytest.mark.asyncio
async def test_aad_session_backend_no_aad_id():

    aad_session_backend = AadSessionBackend()
    request = namedtuple("request", ["session"])
    request.session = {}
    request.session["something"] = "somewhere"

    cred, user = await aad_session_backend.authenticate(request=request)

    assert isinstance(cred, AuthCredentials)
    assert isinstance(user, UnauthenticatedUser)


@pytest.mark.asyncio
async def test_aad_session_backend_no_token_cache():

    aad_session_backend = AadSessionBackend()
    request = namedtuple("request", ["session"])
    request.session = {}
    request.session["aad_id"] = "123456"

    cred, user = await aad_session_backend.authenticate(request=request)

    assert isinstance(cred, AuthCredentials)
    assert isinstance(user, UnauthenticatedUser)


@pytest.mark.asyncio
async def test_aad_session_backend_token_cache(client_credential):

    aad_session_backend = AadSessionBackend()
    request = namedtuple("request", ["session"])
    request.session = {}
    request.session["aad_id"] = "123456"

    options = AzureAdSettings()
    scopes_identifiers = options.api_scopes_identifiers

    payload = gen_payload(
        options, client_credential.private_key, scp=scopes_identifiers
    )

    token = json.loads(payload)
    access_token = token["access_token"]
    id_token = token["access_token"]

    token_cache = (
        '{\n "AccessToken": '
        '  {\n "id": '
        '    {\n  "credential_type": "AccessToken", '
        '        \n  "secret": "' + access_token + '",'
        '        \n  "home_account_id": "A",'
        '        \n  "environment": "login.microsoftonline.com", '
        '        \n  "client_id": "client_id", '
        '        \n  "target": "https://contoso.com/client_id/user_impersonation", '
        '        \n  "realm": "1ca8bd94-3c97-4fc6-8955-bad266b43f0b", '
        '        \n  "token_type": "Bearer", '
        '        \n  "cached_at": "1631724977", '
        '        \n  "expires_on": "1631728575", '
        '        \n  "extended_expires_on": "1631728575" '
        "        \n        }\n    }, "
        '  \n "Account": '
        '    { \n "account_id": '
        '      {\n "home_account_id": "a-a", '
        '        \n "environment": "login.microsoftonline.com",'
        '        \n "realm": "1ca8bd94-3c97-4fc6-8955-bad266b43f0b", '
        '        \n "local_account_id": "c6b74d2d-f41e-4bc4-8cdf-29b6c28d4038", '
        '        \n "username": "johndoe@contoso.com",'
        '        \n "authority_type": "MSSTS" '
        "        \n   }\n    }, "
        '   \n "IdToken": '
        '      {\n "id": '
        '        {\n "credential_type": "IdToken", '
        '         \n "secret": "' + id_token + '", '
        '         \n "home_account_id": "A", '
        '         \n "environment": "login.microsoftonline.com", '
        '         \n "realm": "1ca8bd94-3c97-4fc6-8955-bad266b43f0b", '
        '         \n "client_id": "client_id" '
        "         \n        }\n    },"
        '    \n "RefreshToken": '
        '       {\n "id": '
        '        {\n "credential_type": "RefreshToken", '
        '         \n "secret": "' + id_token + '", '
        '         \n "home_account_id": "A", '
        '         \n "environment": "login.microsoftonline.com", '
        '         \n "client_id": "client_id", '
        '         \n "target": "https://contoso.com/client_id/user_impersonation", '
        '         \n "last_modification_time": "1631724977"\n        }\n    },'
        '    \n "AppMetadata": '
        '       {\n  "appmetadata-id": '
        '        {\n  "client_id": "client_id", '
        '         \n  "environment": "login.microsoftonline.com"\n        }\n    }\n}'
    )

    cache_session = await CacheManager.get(request.session["aad_id"])
    cache_session["token_cache"] = token_cache
    await CacheManager.set(request.session["aad_id"], cache_session)

    cred, user = await aad_session_backend.authenticate(request=request)

    assert isinstance(cred, AuthCredentials)
    assert isinstance(user, AadUser)



from fastapi.security.utils import get_authorization_scheme_param
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
)
from starlette.requests import Request

from .aad_autherror import AuthError
from .aad_helpers import _validate_claims, ensure_user_from_token
from .aad_options import AzureAdSettings
from .aad_token import AuthToken


class AadBearerBackend(AuthenticationBackend):

    # static json, containing discovery urls
    _discovery_keys = None

    def __init__(self, options: AzureAdSettings = None, env_path: str = None, **kwargs):

        if options is None:
            options = AzureAdSettings(_env_file=env_path)

        self.options = options

        # options to decode bearer with another certificate then azure
        # (useful for Tests)
        self.public_key = kwargs.pop("public_key", None)

    # Implementation for middleware for scarlette AuthenticationBackend
    async def authenticate(self, request):
        """Authenticate a request.
        If authentication is successful, defining a user instance
        """

        if "Authorization" not in request.headers:
            return

        try:
            # get token from header
            auth_token = self.get_token_from_header(request)

            # decode token
            user = ensure_user_from_token(auth_token, public_key=self.public_key)

            # validate the claims
            _validate_claims(
                user.claims,
                audiences=self.options.audiences,
                issuers=self.options.get_available_issuers(),
            )
            return AuthCredentials(user.scopes), user

        except Exception as ex:
            raise AuthenticationError("Invalid auth credentials", ex)

    @staticmethod
    def get_token_from_header(request: Request):
        """Get the bearer token from the request header."""

        # get authorization header
        authorization = request.headers.get("Authorization")

        # If not bearer, return None
        if not authorization:
            raise AuthError(
                "authorization_header_missing", "Authorization header is expected"
            )

        # get schema from header
        scheme, bearer = get_authorization_scheme_param(authorization)

        if scheme.lower() != "bearer":
            raise AuthError(
                "authorization_invalid_header",
                "Authorization header must start with" " Bearer",
            )

        if not bearer:
            raise AuthError(
                "authorization_token_not_found",
                "Token not found in the authorization header",
            )
        return AuthToken(bearer)



import json
from datetime import datetime, timedelta
from typing import List

from authlib.common import encoding
from authlib.jose import jwt

from aad.aad_options import AzureAdSettings


def gen_client_info(options: AzureAdSettings, **kwargs):
    """
    Generate a fake client_info payload to ensure cache will work correctly
    """

    uid = kwargs.pop("uid", "jdoe1")
    utid = kwargs.pop("utid", options.tenant_id)

    header = {"uid": uid, "utid": utid}

    dump = encoding.json_dumps(header)
    dump_bytes = encoding.json_b64encode(dump)

    return dump_bytes.decode("utf-8")


def gen_access_token(
    options: AzureAdSettings, private_key: str, use_service_principal=False, **kwargs
):
    """
    Generate a fake token to ensure the validation is correctly done on server side
     with a valid token
    """

    # Generate a due date for the token
    now = datetime.now()
    iat = kwargs.pop("iat", int(datetime.timestamp(now)))
    nbf = kwargs.pop("nbf", iat)
    exp = kwargs.pop("exp", int(datetime.timestamp(now + timedelta(minutes=60))))

    # identifier
    oid = kwargs.pop("oid", "jdoe1")
    tid = kwargs.pop("tid", options.tenant_id)

    # Generate audience and issuer to validate token issuer
    client_id = kwargs.pop("client_id", options.client_id)
    tenant_id = kwargs.pop("tenant_id", options.tenant_id)
    domain = kwargs.pop("domain", options.domain)
    iss = kwargs.pop("iss", f"https://sts.windows.net/{tenant_id}/")
    aud = kwargs.pop("aud", f"https://{domain}/{client_id}")
    scp = kwargs.pop("scp", None)

    # Generate payload without user information
    payload = {
        "iss": iss,
        "aud": aud,
        "iat": iat,
        "exp": exp,
        "nbf": nbf,
        "appid": client_id,
        "oid": oid,
        "tid": tid,
    }

    # Generate user payload if token is issued from a user authentication
    if not use_service_principal:
        family_name = kwargs.pop("family_name", "Doe")
        given_name = kwargs.pop("given_name", "John")
        name = f"{given_name} {family_name}"
        roles = kwargs.pop("roles", None)
        email = kwargs.pop("email", f"{given_name}.{family_name}@{domain}")

        payload["unique_name"] = email
        payload["family_name"] = family_name
        payload["given name"] = given_name
        payload["name"] = name

        if scp is not None:
            if not isinstance(scp, List):
                scp = [scp]

            new_scp = []
            for scope in scp:
                scope_array = scope.split("/")
                if len(scope_array) >= 1:
                    scope = scope_array[len(scope_array) - 1]
                new_scp.append(scope)
            scp = new_scp

            payload["scp"] = scp

        if roles is not None:
            payload["roles"] = roles

    header = {"alg": "RS256", "typ": "JWT", "kid": "nOo3ZDrODXEK1jKWhXslHR_KXEg"}

    dump_bytes = jwt.encode(header, payload, private_key)

    return dump_bytes.decode("utf-8")


def gen_payload(
    options: AzureAdSettings, private_key: str, use_service_principal=False, **kwargs
):
    """Generate a fake payload with an access_token and client_info if
    user credentials"""

    access_token = gen_access_token(
        options=options,
        private_key=private_key,
        use_service_principal=use_service_principal,
        **kwargs,
    )

    if use_service_principal:
        return json.dumps({"access_token": access_token})
    else:
        client_info = gen_client_info(options, **kwargs)
        return json.dumps({"access_token": access_token, "client_info": client_info})



import requests


class AadDiscoverKey:

    # static json, containing discovery urls
    _discovery_keys = None

    @staticmethod
    def get_discovery_key_json(key_url: str):

        if AadDiscoverKey._discovery_keys is None:
            AadDiscoverKey._discovery_keys = requests.get(key_url).json()

        return AadDiscoverKey._discovery_keys
